@@ -382,6 +382,8 @@ gweather_location_unref (GWeatherLocation *loc)
 
     if (loc->metar_code_cache)
 	g_hash_table_unref (loc->metar_code_cache);
+    if (loc->timezone_cache)
+	g_hash_table_unref (loc->timezone_cache);
 
     g_slice_free (GWeatherLocation, loc);
 }
@@ -504,6 +506,98 @@ gweather_location_get_children (GWeatherLocation *loc)
 	return &no_children;
 }
 
+static void
+foreach_cities (GWeatherLocation *loc,
+		GFunc             callback,
+		gpointer          user_data)
+{
+    if (loc->level == GWEATHER_LOCATION_CITY) {
+	callback (loc, user_data);
+    } else if (loc->children) {
+	GWeatherLocation *child;
+	for (child = *loc->children; child != NULL; child++)
+	    foreach_cities (child, callback, user_data);
+    }
+}
+
+struct FindNearestCityData {
+    double latitude;
+    double longitude;
+    GWeatherLocation *location;
+    double distance;
+};
+
+static double
+location_distance (double lat1, double long1,
+		   double lat2, double long2)
+{
+    /* average radius of the earth in km */
+    static const double radius = 6372.795;
+
+    return acos (cos (lat1) * cos (lat2) * cos (long1 - long2) +
+		 sin (lat1) * sin (lat2)) * radius;
+}
+
+static void
+find_nearest_city (GWeatherLocation *location,
+		   gpointer          user_data) {
+    struct FindNearestCityData *data = user_data;
+
+    double distance = location_distance (location->latitude, location->longitude,
+					 data->latitude, data->longitude);
+
+    if (data->location == NULL || data->distance < distance) {
+	data->location = location;
+	data->distance = distance;
+    }
+}
+
+/**
+ * gweather_location_find_nearest_city:
+ * @loc: (allow-none): The parent location, which will be searched recursively
+ * @lat: Latitude, in degrees
+ * @lon: Longitude, in degrees
+ *
+ * Finds the nearest city to the passed latitude and
+ * longitude, among the descendants of @loc.
+ *
+ * @loc must be at most a %GWEATHER_LOCATION_LEVEL_ADM2 location.
+ * This restriction may be lifted in a future version.
+ *
+ * Note that this function does not check if (@lat, @lon) fall inside
+ * @loc, or are in the same region and timezone as the return value.
+ *
+ * Returns: (transfer full): the city closest to (@lat, @lon), in the
+ *          region or administrative district of @loc.
+ */
+GWeatherLocation *
+gweather_location_find_nearest_city (GWeatherLocation *loc,
+				     double            lat,
+				     double            lon)
+{
+   /* The data set really isn't too big. Don't concern ourselves
+     * with a proper nearest neighbors search. Instead, just do
+     * an O(n) search. */
+    struct FindNearestCityData data;
+
+    g_return_val_if_fail (loc == NULL || loc->level < GWEATHER_LOCATION_CITY, NULL);
+
+    if (loc == NULL)
+	loc = gweather_location_get_world ();
+
+    lat = lat * M_PI / 180.0;
+    lon = lon * M_PI / 180.0;
+
+    data.latitude = lat;
+    data.longitude = lon;
+    data.location = NULL;
+    data.distance = 0.0;
+
+    foreach_cities (loc, (GFunc) find_nearest_city, &data);
+
+    return gweather_location_ref (data.location);
+}
+
 /**
  * gweather_location_has_coords:
  * @loc: a #GWeatherLocation
@@ -553,17 +647,14 @@ gweather_location_get_coords (GWeatherLocation *loc,
 double
 gweather_location_get_distance (GWeatherLocation *loc, GWeatherLocation *loc2)
 {
-    /* average radius of the earth in km */
-    static const double radius = 6372.795;
-
     g_return_val_if_fail (loc != NULL, 0);
     g_return_val_if_fail (loc2 != NULL, 0);
 
     //g_return_val_if_fail (loc->latlon_valid, 0.0);
     //g_return_val_if_fail (loc2->latlon_valid, 0.0);
 
-    return acos (cos (loc->latitude) * cos (loc2->latitude) * cos (loc->longitude - loc2->longitude) +
-		 sin (loc->latitude) * sin (loc2->latitude)) * radius;
+    return location_distance (loc->latitude, loc->longitude,
+			      loc2->latitude, loc2->longitude);
 }
 
 /**
@@ -714,7 +805,7 @@ gweather_location_get_code (GWeatherLocation *loc)
  * calling gweather_location_get_name() on the location's parent. For
  * other locations it will return %NULL.
  *
- * Return value: (allow-none) @loc's city name, or %NULL
+ * Return value: (allow-none): @loc's city name, or %NULL
  **/
 char *
 gweather_location_get_city_name (GWeatherLocation *loc)
@@ -904,8 +995,8 @@ gweather_location_format_one_serialize (GWeatherLocation *location)
 			  location->latitude,
 			  location->longitude,
 			  location->parent && location->parent->latlon_valid,
-			  location->parent ? location->parent->latitude : 0.0d,
-			  location->parent ? location->parent->longitude : 0.0d);
+			  location->parent ? location->parent->latitude : 0.0,
+			  location->parent ? location->parent->longitude : 0.0);
 }
 
 GWeatherLocation *
